/*
 * Westell DSL modem monitor program
 * 
 * by David Johnson
 * Copyright (C) 2005-2009 David Johnson
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */

#define _BSD_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <sys/time.h>
#include <time.h>

enum wt_status_e
{
  WT_STATUS_INITIALIZING,
  WT_STATUS_NO_CARRIER,
  WT_STATUS_TRAINING,
  WT_STATUS_ONLINE,
} __attribute__((__packed__));

typedef uint16_t wt_counter16_t;
typedef uint32_t wt_counter32_t;

struct wt_direction_state_t
{
  int16_t   margin;
  uint16_t  power_output;
  uint16_t  attenuation;
  uint16_t  data_rate;
} __attribute__((__packed__));

struct wt_atm_state_t
{
  wt_counter32_t               fec_errors;
  wt_counter32_t               crc_errors;
  wt_counter32_t               hec_errors;
  wt_counter32_t               signal_lost;
  wt_counter32_t               frame_lost;
  wt_counter32_t               tx_user;
  wt_counter32_t               rx_user;
  uint8_t                      __unknown[8];
  wt_counter32_t               loopback;
} __attribute__((__packed__));

#define DATA_MASK_ATM_FEC_ERRORS               (0x0000FFFFU)
#define DATA_MASK_ATM_CRC_ERRORS               (0x0000FFFFU) /* ? */
#define DATA_MASK_ATM_HEC_ERRORS               (0x0000FFFFU) /* ? */
#define DATA_MASK_ATM_SIGNAL_LOST              (0x0000FFFFU) /* ? */
#define DATA_MASK_ATM_FRAME_LOST               (0x0000FFFFU) /* ? */
#define DATA_MASK_ATM_TX_USER                  (0x00FFFFFFU)
#define DATA_MASK_ATM_RX_USER                  (0x00FFFFFFU)
#define DATA_MASK_ATM_LOOPBACK                 (0x0000FFFFU) /* ? */

struct wt_eth_state_t
{
  wt_counter32_t               tx_packets;
  wt_counter32_t               rx_packets;
  wt_counter32_t               discard;
} __attribute__((__packed__));

#define DATA_MASK_ETH_TX_PACKETS               (0xFFFFFFFFU) /* ? */
#define DATA_MASK_ETH_RX_PACKETS               (0xFFFFFFFFU) /* ? */
#define DATA_MASK_ETH_DISCARD                  (0x00FFFFFFU) /* ? */

#define DATA_TOTAL_SIZE                        (120)

struct wt_notify_data_t
{
  uint8_t                      magic[3];                 /*  3 bytes */
  enum wt_status_e             status;                   /*  1 byte  */
  uint32_t                     time;                     /*  4 bytes */
  struct wt_direction_state_t  up;                       /*  8 bytes */
  struct wt_direction_state_t  down;                     /*  8 bytes */
  struct wt_atm_state_t        atm;                      /* 40 bytes */
  uint8_t                      __unknown1[34];           /* 34 bytes */
  struct wt_eth_state_t        eth;                      /* 12 bytes */
  uint8_t                      __unknown2[10];           /* 10 bytes */
} __attribute__((__packed__));

#if __BYTE_ORDER == __BIG_ENDIAN
#define LE_TO_CPU16(from) ({ uint16_t x = from; x = ((x & 0x00FFU) << 8) | ((x & 0xFF00U) >> 8); x; })
#define LE_TO_CPU32(from) ({ uint32_t x = from; x = ((x & 0x000000FFU) << 24) | ((x & 0x0000FF00U) << 8) | ((x & 0x00FF0000U) >> 8) | ((x & 0xFF000000U) >> 24); x; })
#elif __BYTE_ORDER == __LITTLE_ENDIAN
#define LE_TO_CPU16(from) ({ from; })
#define LE_TO_CPU32(from) ({ from; })
#else
#error "what endian?"
#endif

#define WESTELL_UDP_CONTROL_PORT   (2420)
#define WESTELL_UDP_NOTIFY_PORT    (1875)
#define WESTELL_MULITCAST_GROUP    "224.73.193.62"

#define BUFFER_SIZE                (256)

static char buffer_s[BUFFER_SIZE];
static char buffer_r[BUFFER_SIZE];
static char buffer_m[BUFFER_SIZE];

static int verbose = 0;
static int got_sigint = 0;

static int double_counts_nonunicast = 1;

static const char *datefmt = "%Y-%m-%d-%H:%M:%S";
static const char *ifs = ",";

#define MAX_CLIENTS                (16)

static int client_fds[MAX_CLIENTS] = { -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1 };


/*
 * handle diff of any data size (by mask) and look for counter resets
 * due to excessive jumps
 */
#define DO_COUNTER_DIFF(elem, ELEM)                     \
  ({                                                    \
    unsigned int _diff = data->elem - lastdata->elem;   \
    _diff &= DATA_MASK_ ## ELEM;                        \
    if ((data->elem < (DATA_MASK_ ## ELEM >> 1)) &&     \
        (_diff > (DATA_MASK_ ## ELEM >> 1)))            \
      {                                                 \
        /* reset */                                     \
        _diff = 0;                                      \
      }                                                 \
    _diff;                                              \
  })

#define ATM_CELL_HEADER_SIZE		(5)
#define ATM_CELL_DATA_SIZE		(48)
#define ATM_CELL_SIZE			(ATM_CELL_HEADER_SIZE+ATM_CELL_DATA_SIZE)
#define LLCSNAP_HEADER_SIZE		(8)
#define ETH_HEADER_SIZE			(14)
#define AAL5_TRAILER_SIZE		(8)
#define MIN_IP_PACKET_SIZE		(20)

/* RFC2684 Bridged Ethernet, assumes no FCS and no PPPoE */
#define TOTAL_PACKET_OVERHEAD		(LLCSNAP_HEADER_SIZE+2+ETH_HEADER_SIZE+AAL5_TRAILER_SIZE)


#define IP_SIZE_TO_CELLS(size)		((int)(((size)+TOTAL_PACKET_OVERHEAD+ATM_CELL_DATA_SIZE-1)/ATM_CELL_DATA_SIZE))

#define DATA_RATE_TO_CPS(rate) ((float)(rate) * 1000 / 8 / ATM_CELL_SIZE)
#define CPS_TO_DATA_RATE(cps) ((float)(cps) * ATM_CELL_SIZE * 8 / 1000)

#define CPS_PPS_TO_IP_DATA_RATE(cps, pps)               \
  ( ( ((float)(cps) * ATM_CELL_DATA_SIZE * 8) - ((float)(pps) * (TOTAL_PACKET_OVERHEAD+ATM_CELL_DATA_SIZE/2) * 8) ) / 1000 )

#define CPS_PPS_TO_IP_DATA_RATE_MAX_ERROR(cps, pps)     \
  ( ((float)(pps) * (ATM_CELL_DATA_SIZE/2) * 8) / 1000 )

enum westell_option_types
{
  OPT_VERBOSE = 1,
  OPT_IFACE,
  OPT_BCAST,
  OPT_RATE,
  OPT_DAEMON,
  OPT_DATEFMT,
  OPT_IFS,
  OPT_ACTION_INFO,
  OPT_ACTION_ENABLE,
  OPT_ACTION_DISABLE,
  OPT_ACTION_MONITOR,
  OPT_ACTION_SERVE,
};

static struct option westell_long_options[] = {
  { "verbose",     0, NULL, OPT_VERBOSE },
  { "iface",       1, NULL, OPT_IFACE },
  { "bcast",       1, NULL, OPT_BCAST },
  { "rate",        1, NULL, OPT_RATE },
  { "daemon",      0, NULL, OPT_DAEMON },
  { "datefmt",     1, NULL, OPT_DATEFMT },
  { "ifs",         1, NULL, OPT_IFS },
  { "info",        0, NULL, OPT_ACTION_INFO },
  { "enable",      0, NULL, OPT_ACTION_ENABLE },
  { "disable",     0, NULL, OPT_ACTION_DISABLE },
  { "monitor",     0, NULL, OPT_ACTION_MONITOR },
  { "serve",       1, NULL, OPT_ACTION_SERVE },
  { NULL,          0, NULL, 0 }
};


static void __attribute__ ((noreturn)) usage(const char *progname)
{
  fprintf(stderr, "Usage: %s [--verbose] --bcast <bcast> [--info] [--enable] [--disable]\n", progname);
  fprintf(stderr, "Usage: %s [--verbose] [--iface <iface>] [--datefmt fmt] [--rate sec] --monitor\n", progname);
  fprintf(stderr, "Usage: %s [--verbose] [--iface <iface>] [--datefmt fmt] [--ifs ifs] [--rate sec] [--daemon] --serve ip:port\n", progname);
  exit(1);
}

static void westell_signal_handler(int signo, siginfo_t *sinfo, void *unused)
{
  switch(signo)
    {
    case SIGALRM:
    case SIGPIPE:
      /* ignore */
      break;
      
    case SIGINT:
      got_sigint = 1;
      break;
      
    default:
      return;
    }
  
  return;
}

static int is_fd_writeable(int fd)
{
  int ret;
  fd_set writefds;
  struct timeval timeout;
  
  FD_ZERO(&writefds);
  FD_SET(fd, &writefds);
  
  timeout.tv_sec = 0;
  timeout.tv_usec = 0;
  
  ret = select(fd+1, NULL, &writefds, NULL, &timeout);
  
  if (ret < 0)
    {
      return 0;
    }
  
  if (FD_ISSET(fd, &writefds))
    {
      return 1;
    }
  
  return 0;
}

static const char *wt_status_to_string(enum wt_status_e status)
{
  switch (status)
    {
    case WT_STATUS_INITIALIZING:
      return "initializing";
      
    case WT_STATUS_NO_CARRIER:
      return "no-carrier";
      
    case WT_STATUS_TRAINING:
      return "training";
      
    case WT_STATUS_ONLINE:
      return "online";
      
    default:
      return "unknown";
    }
}

static int do_transaction(const int fd,
                          const struct sockaddr_in *remote,
                          const char *send_buf,
                          const ssize_t send_len,
                          char *recv_buf,
                          ssize_t *recv_len,
                          const char *magic_buf,
                          const ssize_t magic_len)
{
  ssize_t call_len;
  int retry = 3;
  int saved_errno;
  
 send_again:
  if (!send_len)
    {
      goto recv_again;
    }
  
  call_len = sendto(fd, send_buf, send_len, 0,
                    (const struct sockaddr*)remote,
                    sizeof(*remote));
  
  if (call_len != send_len)
    {
      perror("send");
      return -1;
    }
  
 recv_again:
  alarm(2);
  call_len = recv(fd, recv_buf, *recv_len, 0);
  saved_errno = errno;
  alarm(0);
  errno = saved_errno;
  
  if ((call_len < 0) && (errno == EINTR))
    {
      if (got_sigint)
        {
          fprintf(stderr,"Interrupted\n");
          return -1;
        }
      if (--retry)
        {
          goto send_again;
        }
      fprintf(stderr,"No reponse\n");
      return -1;
    }
  else if (call_len <= 0)
    {
      perror("recv");
      return -1;
    }
  
  if (magic_len && (call_len < sizeof(magic_len)))
    {
      goto recv_again;
    }
  
  if (magic_len && (memcmp(recv_buf, magic_buf, magic_len)))
    {
      goto recv_again;
    }
  
  *recv_len = call_len;
  return 0;
}


static int do_info(struct sockaddr_in *opt_bcast)
{
  int fd;
  int tmp = 1;
  ssize_t recv_len;
  int ll;
  
  fd = socket(PF_INET, SOCK_DGRAM, 0);
  if (fd < 0)
    {
      perror("socket");
      return -1;
    }
  
  if (setsockopt(fd, SOL_SOCKET, SO_BROADCAST, &tmp, sizeof(tmp)) < 0)
    {
      perror("SO_BROADCAST");
      goto error_out;
    }
  
  opt_bcast->sin_port = htons(WESTELL_UDP_CONTROL_PORT);
  
  /* get serial and model numbers */
  
  printf("Obtaining Serial and Model Numbers:\n");
  
  buffer_s[0] = 0x00;
  buffer_s[1] = 0x0c;
  buffer_s[2] = 0x00;
  
  buffer_m[0] = 0x00;
  buffer_m[1] = 0x03;
  buffer_m[2] = 0x00;
  
  recv_len = BUFFER_SIZE;
  if (do_transaction(fd, opt_bcast, buffer_s, 3, buffer_r, &recv_len, buffer_m, 3) < 0)
    {
      goto error_out;
    }

  for (ll = recv_len-4; ll > 0; ll--)
    {
      ll -= printf("%s\n", buffer_r+recv_len-ll) - 1;
    }
  
  
  /* get software version */
  
  printf("\nObtaining Software Versions:\n");
  
  buffer_s[0] = 0x00;
  buffer_s[1] = 0x01;
  buffer_s[2] = 0x00;
  
  buffer_m[0] = 0x00;
  buffer_m[1] = 0x03;
  buffer_m[2] = 0x00;
  
  recv_len = BUFFER_SIZE;
  if (do_transaction(fd, opt_bcast, buffer_s, 3, buffer_r, &recv_len, buffer_m, 3) < 0)
    {
      goto error_out;
    }

  for (ll = recv_len-4; ll > 0; ll--)
    {
      ll -= printf("%s\n", buffer_r+recv_len-ll) - 1;
    }
  
  /* get firmware version */
  
  printf("\nObtaining Firmware Versions:\n");
  
  buffer_s[0] = 0x00;
  buffer_s[1] = 0x06;
  
  buffer_m[0] = 0x00;
  buffer_m[1] = 0x03;
  buffer_m[2] = 0x00;
  
  recv_len = BUFFER_SIZE;
  if (do_transaction(fd, opt_bcast, buffer_s, 2, buffer_r, &recv_len, buffer_m, 3) < 0)
    {
      goto error_out;
    }

  for (ll = recv_len-4; ll > 0; ll--)
    {
      ll -= printf("%s\n", buffer_r+recv_len-ll) - 1;
    }
  
  close(fd);
  return 0;
  
 error_out:
  close(fd);
  return -1;
}


static int do_enable(struct sockaddr_in *opt_bcast)
{
  int fd;
  int tmp = 1;
  ssize_t recv_len;
  
  fd = socket(PF_INET, SOCK_DGRAM, 0);
  if (fd < 0)
    {
      perror("socket");
      return -1;
    }
  
  if (setsockopt(fd, SOL_SOCKET, SO_BROADCAST, &tmp, sizeof(tmp)) < 0)
    {
      perror("SO_BROADCAST");
      goto error_out;
    }
  
  if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &tmp, sizeof(tmp)) < 0)
    {
      perror("SO_REUSEADDR");
      goto error_out;
    }
  
  opt_bcast->sin_port = htons(WESTELL_UDP_NOTIFY_PORT);
  
  /* get serial and model numbers */
  
  printf("Enabling Notification\n");
  
  buffer_s[0] = 0x00;
  buffer_s[1] = 0x30;
  buffer_s[2] = 0x01;
  buffer_s[3] = 0xe0;
  buffer_s[4] = 0x49;
  buffer_s[5] = 0xc1;
  buffer_s[6] = 0x3e;
  
  buffer_m[0] = 0x00;
  
  recv_len = BUFFER_SIZE;
  if (do_transaction(fd, opt_bcast, buffer_s, 7, buffer_r, &recv_len, buffer_m, 1) < 0)
    {
      goto error_out;
    }

  printf("Enabled\n");
  
  close(fd);
  return 0;
  
 error_out:
  close(fd);
  return -1;
}

static void print_direction(int duration_ms,
                            const struct wt_direction_state_t *lastdata,
                            const struct wt_direction_state_t *data,
                            const char *desc)
{
  float max_cps;
  
  printf("  %s\n", desc);

#define PRINT_DIRECTION_ITEM(desc, units, elem)                 \
  do {                                                          \
    float diff = (float)data->elem - (float)lastdata->elem;     \
    printf("    %-18s : %4.1f %-5s",                            \
           desc,                                                \
           (float)data->elem / 10, units);                      \
    if (diff)                                                   \
      {                                                         \
        printf("    [ %5.1f %-3s]\n", diff / 10, units);        \
      }                                                         \
    else                                                        \
      {                                                         \
        printf("\n");                                           \
      }                                                         \
  } while(0)

  PRINT_DIRECTION_ITEM("Margin",             "db",    margin);
  PRINT_DIRECTION_ITEM("Power",              "db/Hz", power_output);
  PRINT_DIRECTION_ITEM("Attenuation",        "db",    attenuation);
  
  max_cps = DATA_RATE_TO_CPS(data->data_rate);
  
  printf("    Max Rate           : %u kbps, %.1f cells/sec\n",
         data->data_rate,
         max_cps);
  
  /* AAL5 rate is rate without ATM header */
  /* 1500 byte IP + 32bytes of overhead (AAL5, LLC/SNAP, Bridge) = 32 cells */
  /*   64 byte IP + 32bytes of overhead (AAL5, LLC/SNAP, Bridge) = 2 cells */
  
  printf("    Max Rate           : %u ATM, %.0f AAL5, %.0f IP-1500, %.0f IP-64\n",
         data->data_rate,
         ATM_CELL_DATA_SIZE * 8 * max_cps / 1000,
         1500 * 8 * max_cps / IP_SIZE_TO_CELLS(1500) / 1000,
         64 * 8 * max_cps / IP_SIZE_TO_CELLS(64) / 1000);
  
  return;
}

static void print_atm(int duration_ms,
                      const struct wt_notify_data_t *lastdata,
                      const struct wt_notify_data_t *data)
{
  printf("  ATM\n");
  
#define PRINT_ATM_ITEM(desc, elem, ELEM)                                \
  do {                                                                  \
    unsigned int diff = DO_COUNTER_DIFF(atm.elem, ATM_ ## ELEM);        \
    printf("    %-18s : %10u", desc, data->atm.elem);                   \
    if (diff)                                                           \
      {                                                                 \
        printf("    [ %6u cells ]\n", diff);                            \
      }                                                                 \
    else                                                                \
      {                                                                 \
        printf("\n");                                                   \
      }                                                                 \
  } while(0)
  
  PRINT_ATM_ITEM("Signal Lost",     signal_lost,       SIGNAL_LOST);
  PRINT_ATM_ITEM("Frame Lost",      frame_lost,        FRAME_LOST);
  PRINT_ATM_ITEM("FEC Errors",      fec_errors,        FEC_ERRORS);
  PRINT_ATM_ITEM("CRC Errors",      crc_errors,        CRC_ERRORS);
  PRINT_ATM_ITEM("HEC Errors",      hec_errors,        HEC_ERRORS);
  PRINT_ATM_ITEM("Loopback Cells",  loopback,          LOOPBACK);
  PRINT_ATM_ITEM("TX Cells",        tx_user,           TX_USER);
  PRINT_ATM_ITEM("RX Cells",        rx_user,           RX_USER);

  if (verbose)
    {
      int ii;
      
      printf("  Unknown              : ");
      for (ii=0; ii<sizeof(data->atm.__unknown); ii++)
        {
          printf("%02x", data->atm.__unknown[ii]);
        }
      printf("\n");
    }
  
  return;
}


static void print_eth(int duration_ms,
                      const struct wt_notify_data_t *lastdata,
                      const struct wt_notify_data_t *data)
{
  printf("  ETH\n");
  
#define PRINT_ETH_ITEM(desc, elem, ELEM)                                \
  do {                                                                  \
    unsigned int diff = DO_COUNTER_DIFF(eth.elem, ETH_ ## ELEM);        \
    printf("    %-18s : %10u",                                          \
           desc,                                                        \
           data->eth.elem);                                             \
    if (diff)                                                           \
      {                                                                 \
        printf("    [ %6u pkts  ]\n", diff);                            \
      }                                                                 \
    else                                                                \
      {                                                                 \
        printf("\n");                                                   \
      }                                                                 \
  } while(0)
  
  PRINT_ETH_ITEM("Dropped Packets", discard,        DISCARD);
  PRINT_ETH_ITEM("TX Packets",      tx_packets,     TX_PACKETS);
  PRINT_ETH_ITEM("RX Packets",      rx_packets,     RX_PACKETS);
  
  return;
}

static void print_usage(int duration_ms,
                        int isup,
                        const struct wt_notify_data_t *lastdata,
                        const struct wt_notify_data_t *data)
{
  const char *desc;
  float used_cps;
  float used_pps;
  float max_cps;
  
  if (isup)
    {
      desc         = "Upstream";
      used_pps     = DO_COUNTER_DIFF(eth.rx_packets, ETH_RX_PACKETS);
      used_cps     = DO_COUNTER_DIFF(atm.tx_user,    ATM_TX_USER);
      max_cps      = DATA_RATE_TO_CPS(data->up.data_rate);
    }
  else
    {
      desc         = "Downstream";
      used_pps     = DO_COUNTER_DIFF(eth.tx_packets, ETH_TX_PACKETS);
      used_cps     = DO_COUNTER_DIFF(atm.rx_user,    ATM_RX_USER);
      max_cps      = DATA_RATE_TO_CPS(data->down.data_rate);
    }
  
  used_pps /= (float)duration_ms / 1000.0;
  used_cps /= (float)duration_ms / 1000.0;
  
  printf("  %s\n", desc);

  printf("    Max Rate           : %6.1f kbps, %6.1f cps\n",
         CPS_TO_DATA_RATE(max_cps),
         max_cps);
  
  if (used_pps)
    { 
      float max_atm_bytes_sec = CPS_TO_DATA_RATE(max_cps);
      float used_atm_bytes_sec = CPS_TO_DATA_RATE(used_cps);
      float used_ip_bytes_sec  = CPS_PPS_TO_IP_DATA_RATE(used_cps, used_pps);
      
      printf("    Current Rate       : %6.1f kbps, %6.1f cps, %6.1f pps      [ %5.1f%% ]\n",
             CPS_TO_DATA_RATE(used_cps),
             used_cps,
             used_pps,
             100.0 * used_cps / max_cps);
      
      max_atm_bytes_sec    *= 1000 / 8;
      used_atm_bytes_sec   *= 1000 / 8;
      used_ip_bytes_sec    *= 1000 / 8;
      
      printf("    IP Averages        : %6.1f bytes/pkt\n",
             used_ip_bytes_sec / used_pps);
      
      printf("    ATM Averages       : %6.1f bytes/pkt, %6.1f cells/pkt\n",
             used_atm_bytes_sec / used_pps,
             used_cps / used_pps);
      
      printf("    ATM Overhead       : %6.1f bytes/pkt, %6.1f bytes/cell, %4.1f%% efficient\n",
             (used_atm_bytes_sec - used_ip_bytes_sec) / used_pps,
             (used_atm_bytes_sec - used_ip_bytes_sec) / used_cps,
             100.0 * (used_ip_bytes_sec / used_atm_bytes_sec));
      
      printf("    Utilization        : %5.1f%% data, %5.1f%% overhead, %5.1f%% idle\n",
             100.0 * (used_ip_bytes_sec) / max_atm_bytes_sec,
             100.0 * ((used_atm_bytes_sec - used_ip_bytes_sec)) / max_atm_bytes_sec,
             100.0 * ((max_atm_bytes_sec - used_atm_bytes_sec)) / max_atm_bytes_sec);
      printf("    Utilization (kbps) : %6.1f data, %6.1f overhead, %6.1f idle\n",
             used_ip_bytes_sec / 1000 * 8,
             (used_atm_bytes_sec - used_ip_bytes_sec) / 1000 * 8,
             (max_atm_bytes_sec - used_atm_bytes_sec) / 1000 * 8);
    }
  else
    {
      printf("    Current Rate       :\n");
      printf("    IP Averages        :\n");
      printf("    ATM Averages       :\n");
      printf("    ATM Overhead       :\n");
      printf("    Utilization        :\n");
      printf("    Utilization (kbps) :\n");
    }
  
  return;
}

static void print_notification(const struct timeval *last,
                               const struct timeval *now,
                               const struct wt_notify_data_t *lastdata,
                               const struct wt_notify_data_t *data)
{
  char datestr[BUFFER_SIZE];
  size_t len;
  struct timeval duration;
  int duration_ms;
  time_t boottime;

  timersub(now, last, &duration);
  duration_ms = duration.tv_sec * 1000 + (duration.tv_usec + 500) / 1000;
  
  len = strftime(datestr, BUFFER_SIZE, datefmt, localtime(&now->tv_sec));
  if (!len || (len == BUFFER_SIZE))
    {
      datestr[0] = '\0';
    }
  
  printf("Got packet on %s:\n", datestr);
  
  if (verbose)
    {
      int ii;
      
      printf("  Magic                : ");
      for (ii=0; ii<sizeof(data->magic); ii++)
        {
          printf("%02x", data->magic[ii]);
        }
      printf("\n");
    }
  
  printf("  Status               : %s\n", wt_status_to_string(data->status));

  /* time field is either uptime since boot or time since epoc */
  boottime = now->tv_sec - data->time;
  if (boottime > 60*60)
    {
      len = strftime(datestr, BUFFER_SIZE, datefmt, localtime(&boottime));
      if (!len || (len == BUFFER_SIZE))
        {
          datestr[0] = '\0';
        }
      printf("  Uptime               : %uD %uH %uM (since %s)\n",
             (data->time / 60 / 60 / 24),
             (data->time / 60 / 60) % 24,
             (data->time / 60) % 60,
             datestr);
    }
  else
    {
      boottime = data->time;
      len = strftime(datestr, BUFFER_SIZE, datefmt, localtime(&boottime));
      if (!len || (len == BUFFER_SIZE))
        {
          datestr[0] = '\0';
        }
      printf("  Time                 : %s (%d sec diff)\n",
             datestr,
             (int)(data->time - now->tv_sec));
    }

  if (data->status == WT_STATUS_ONLINE)
    {
      print_direction(duration_ms, &lastdata->up,   &data->up,   "Upstream");
      print_direction(duration_ms, &lastdata->down, &data->down, "Downstream");
    }
  
  print_atm(duration_ms, lastdata, data);
  print_eth(duration_ms, lastdata, data);
  
  if (data->status == WT_STATUS_ONLINE)
    {
      print_usage(duration_ms, 1, lastdata, data);
      print_usage(duration_ms, 0, lastdata, data);
    }

  if (verbose)
    {
      int ii;
      
      printf("  Unknown1             : ");
      for (ii=0; ii<sizeof(data->__unknown1); ii++)
        {
          printf("%02x", data->__unknown1[ii]);
        }
      printf("\n");
      
      printf("  Unknown2             : ");
      for (ii=0; ii<sizeof(data->__unknown2); ii++)
        {
          printf("%02x", data->__unknown2[ii]);
        }
      printf("\n");
    }
}

static void process_notification(unsigned int opt_rate,
                                 const char *buffer, ssize_t buffer_len,
                                 void (*callback)(const struct timeval *last,
                                                  const struct timeval *now,
                                                  const struct wt_notify_data_t *lastdata,
                                                  const struct wt_notify_data_t *data))
{
  static struct wt_notify_data_t lastdatad = {};
  static struct timeval last;
  struct wt_notify_data_t datad;

  struct wt_notify_data_t *lastdata = &lastdatad;
  struct wt_notify_data_t *data = &datad;

  struct timeval duration;
  int duration_ms;

  struct timeval now;
  gettimeofday(&now, NULL);
  
  /* verify packet */
  if (buffer_len != DATA_TOTAL_SIZE)
    {
      printf("Ignoring bad packet\n");
    }
  memcpy(data, buffer, sizeof(*data));
  
  /* packet data is in little endian */
  data->time                   = LE_TO_CPU32(data->time);
  data->up.margin              = LE_TO_CPU16(data->up.margin);
  data->up.power_output        = LE_TO_CPU16(data->up.power_output);
  data->up.attenuation         = LE_TO_CPU16(data->up.attenuation);
  data->up.data_rate           = LE_TO_CPU16(data->up.data_rate);
  data->down.margin            = LE_TO_CPU16(data->down.margin);
  data->down.power_output      = LE_TO_CPU16(data->down.power_output);
  data->down.attenuation       = LE_TO_CPU16(data->down.attenuation);
  data->down.data_rate         = LE_TO_CPU16(data->down.data_rate);
  data->atm.fec_errors         = LE_TO_CPU32(data->atm.fec_errors);
  data->atm.crc_errors         = LE_TO_CPU32(data->atm.crc_errors);
  data->atm.hec_errors         = LE_TO_CPU32(data->atm.hec_errors);
  data->atm.signal_lost        = LE_TO_CPU32(data->atm.signal_lost);
  data->atm.frame_lost         = LE_TO_CPU32(data->atm.frame_lost);
  data->atm.tx_user            = LE_TO_CPU32(data->atm.tx_user);
  data->atm.rx_user            = LE_TO_CPU32(data->atm.rx_user);
  data->atm.loopback           = LE_TO_CPU32(data->atm.loopback);
  data->eth.tx_packets         = LE_TO_CPU32(data->eth.tx_packets);
  data->eth.rx_packets         = LE_TO_CPU32(data->eth.rx_packets);
  data->eth.discard            = LE_TO_CPU32(data->eth.discard);
  
  /* this notification is counted in eth tx counter so compensate */
  lastdata->eth.tx_packets++;
  
  
  /* handle lots of bugs in the westell firmware.... */
  
  if (data->status != WT_STATUS_ONLINE)
    {
      /* if modem isn't online, it sends fixed (non-zero) data in most fields */
      memset(&data->up, 0, sizeof(data->up));
      memset(&data->down, 0, sizeof(data->down));
      memcpy(&lastdata->atm, &data->atm, sizeof(lastdata->atm));
    }
  
  /* sometimes margin is sent negative */
  if (data->down.margin < 0)
    {
      data->down.margin = -data->down.margin;
    }
  
  /*
   * Some modems double count every broadcast and multicast ethernet
   * packet they tx (including this notification).  If effected by this
   * bug, the eth.tx_packets delta will never be zero because this
   * notification packet is a multicast.
   * 
   * If the eth.tx_packets delta is ever zero then we know this modem
   * is not effected by this bug, otherwise we must assume its broken
   * until we do receive an idle sample (if ever).
   */
  if (lastdata->eth.tx_packets == data->eth.tx_packets)
    {
      double_counts_nonunicast = 0;
    }
  else if (double_counts_nonunicast)
    {
      /*
       * Try to correct for bug, but we have no idea how many
       * broadcast/multicast packets were double counted (other than
       * this notification) so handle that one.
       */
      lastdata->eth.tx_packets++; /* this notification (multicast) */
    }
  
  /* when link comes up for the first time various data is bogus */
  if (data->up.margin == 0x7FFF)
    {
      return;
    }
  
  /* sometimes packet is corrupt with rx cells == tx cells */
  if (data->atm.tx_user && (data->atm.tx_user == data->atm.rx_user) &&
      lastdata->atm.tx_user && (lastdata->atm.tx_user != lastdata->atm.rx_user))
    {
      return;
    }
  
  if (double_counts_nonunicast)
    {
      int delta_tx_eth;
      int delta_rx_atm;
      int max_tx_eth;

      /*
       * We can try to guess if the number of double counts is
       * excessive by looking at the cell vs packet counts.
       * 
       * Is the tx eth packet count too large for the number of rx
       * atm cells?  If so correct tx eth packets as best we can.
       */

      delta_rx_atm = DO_COUNTER_DIFF(atm.rx_user, ATM_RX_USER);
      delta_tx_eth = DO_COUNTER_DIFF(eth.tx_packets, ETH_TX_PACKETS);
      max_tx_eth = delta_rx_atm / IP_SIZE_TO_CELLS(MIN_IP_PACKET_SIZE);
      
      if (delta_tx_eth > max_tx_eth)
        {
          lastdata->eth.tx_packets += (delta_tx_eth - max_tx_eth);
        }
    }
  
  /* ok that's alot of bugs... now process packet */
  
  if (!last.tv_sec)
    {
      *lastdata = *data;
      last = now;
    }
  
  timersub(&now, &last, &duration);
  duration_ms = duration.tv_sec * 1000 + (duration.tv_usec + 500) / 1000;
  
  if (duration_ms >= (opt_rate*1000)-10) /* allow 10msec jitter */
    {
      if (callback)
        {
          callback(&last, &now, lastdata, data);
        }
      
      *lastdata = *data;
      last = now;
    }
  
  return;
}


static int do_monitor(int ifindex,
                      unsigned int opt_rate)
{
  int fd;
  int tmp = 1;
  int retval;
  ssize_t recv_len;
  struct sockaddr_in local = {};
  struct ip_mreqn mcast = {};
  
  fd = socket(PF_INET, SOCK_DGRAM, 0);
  if (fd < 0)
    {
      perror("socket");
      return -1;
    }
  
  if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &tmp, sizeof(tmp)) < 0)
    {
      perror("SO_REUSEADDR");
      goto error_out;
    }
  
  local.sin_family = AF_INET;
  local.sin_addr.s_addr = INADDR_ANY;
  local.sin_port = htons(WESTELL_UDP_NOTIFY_PORT);
  
  if (bind(fd, (struct sockaddr *)&local, sizeof(local)) < 0)
    {
      perror("bind");
      retval = -1;
      goto error_out;
    }
  
  inet_aton(WESTELL_MULITCAST_GROUP, &mcast.imr_multiaddr);
  mcast.imr_address.s_addr   = INADDR_ANY;
  mcast.imr_ifindex          = ifindex;
  
  if (setsockopt(fd, SOL_IP, IP_ADD_MEMBERSHIP, &mcast, sizeof(mcast)) < 0)
    {
      perror("IP_ADD_MEMBERSHIP");
      retval = -1;
      goto error_out;
    }
  
  /* get packet */
  
  printf("Waiting for Packet\n");
  
  while (1)
    {
      buffer_m[0] = 0x00;
      buffer_m[1] = 0x32;
      buffer_m[2] = 0x01;
      
      recv_len = BUFFER_SIZE;
      if (do_transaction(fd, NULL, NULL, 0, buffer_r, &recv_len, buffer_m, 3) < 0)
        {
          retval = -1;
          goto error_out_drop;
        }
      
      process_notification(opt_rate, buffer_r, recv_len, print_notification);
    }
  retval = 0;
  
 error_out_drop:
  inet_aton(WESTELL_MULITCAST_GROUP, &mcast.imr_multiaddr);
  mcast.imr_address.s_addr   = INADDR_ANY;
  mcast.imr_ifindex          = ifindex;
  
  if (setsockopt(fd, SOL_IP, IP_DROP_MEMBERSHIP, &mcast, sizeof(mcast)) < 0)
    {
      perror("IP_DROP_MEMBERSHIP");
      retval = -1;
      goto error_out;
    }
  
 error_out:
  close(fd);
  return retval;
}


static void serve_notification(const struct timeval *last,
                               const struct timeval *now,
                               const struct wt_notify_data_t *lastdata,
                               const struct wt_notify_data_t *data)
{
  struct timeval duration;
  int duration_ms;
  int client;
  int len;
  int len2;
  
  timersub(now, last, &duration);
  duration_ms = duration.tv_sec * 1000 + (duration.tv_usec + 500) / 1000;

  len = strftime(buffer_s, BUFFER_SIZE, datefmt, localtime(&now->tv_sec));
  if (!len || (len == BUFFER_SIZE))
    {
      buffer_s[0] = '\0';
      len = 0;
    }
  
  len2 = snprintf(buffer_s+len, BUFFER_SIZE-len,
                  "%s%.3f%s%s%s%u"
                  "%s%.1f%s%.1f%s%.1f%s%u"
                  "%s%.1f%s%.1f%s%.1f%s%u"
                  "%s%u%s%u%s%u%s%u%s%u%s%u%s%u%s%u%s%u%s%u"
                  "%s%u%s%u%s%u"
                  "\n",
                  ifs, (float)duration_ms/1000.0,
                  ifs, wt_status_to_string(data->status),
                  ifs, data->time,
                  
                  ifs, (float)data->up.margin / 10,
                  ifs, (float)data->up.power_output / 10,
                  ifs, (float)data->up.attenuation / 10,
                  ifs, data->up.data_rate,
                  
                  ifs, (float)data->down.margin / 10,
                  ifs, (float)data->down.power_output / 10,
                  ifs, (float)data->down.attenuation / 10,
                  ifs, data->down.data_rate,
                  
                  ifs, DO_COUNTER_DIFF(atm.signal_lost, ATM_SIGNAL_LOST),
                  ifs, DO_COUNTER_DIFF(atm.frame_lost, ATM_FRAME_LOST),
                  ifs, DO_COUNTER_DIFF(atm.fec_errors, ATM_FEC_ERRORS),
                  ifs, DO_COUNTER_DIFF(atm.crc_errors, ATM_CRC_ERRORS),
                  ifs, DO_COUNTER_DIFF(atm.hec_errors, ATM_HEC_ERRORS),
                  ifs, 0, /* reserved */
                  ifs, DO_COUNTER_DIFF(atm.loopback, ATM_LOOPBACK),
                  ifs, 0, /* reserved */
                  ifs, DO_COUNTER_DIFF(atm.tx_user, ATM_TX_USER),
                  ifs, DO_COUNTER_DIFF(atm.rx_user, ATM_RX_USER),
                  
                  ifs, DO_COUNTER_DIFF(eth.discard, ETH_DISCARD),
                  ifs, DO_COUNTER_DIFF(eth.tx_packets, ETH_TX_PACKETS),
                  ifs, DO_COUNTER_DIFF(eth.rx_packets, ETH_RX_PACKETS));
  if ((len2 < 0) || (len2 >= BUFFER_SIZE-len))
    {
      buffer_s[0] = '\n';
      buffer_s[1] = '\0';
      len2 = 1;
    }
  len += len2;
  
  if (verbose) printf("Got Notification\n");
  
  if (verbose > 1)
    {
      print_notification(last, now, lastdata, data);
    }
  
  for(client=0; client<MAX_CLIENTS; client++)
    {
      if (client_fds[client] != -1)
        {
          if (is_fd_writeable(client_fds[client]))
            {
              ssize_t writelen;
              
              if (verbose) printf("Sending Notification to client slot %d\n", client);
              
              writelen = write(client_fds[client],
                               buffer_s, len);
              if (writelen < 0)
                {
                  if (verbose) printf("Lost Client slot %d\n", client);
                  close(client_fds[client]);
                  client_fds[client] = -1;
                }
              /* ignore partial writes */
            }
          else
            {
              if (verbose) printf("Skipping Notification to client slot %d\n", client);
            }
        }
    }
  
  return;
}

static int do_serve(int ifindex,
                    unsigned int opt_rate,
                    int opt_daemon,
                    const char *opt_serve)
{
  int wt_fd;
  int tmp = 1;
  int listen_fd = -1;
  int retval = -1;
  ssize_t recv_len;
  struct sockaddr_in local = {};
  struct ip_mreqn mcast = {};
  
  /* setup udp listen */
  wt_fd = socket(PF_INET, SOCK_DGRAM, 0);
  if (wt_fd < 0)
    {
      perror("socket");
      return -1;
    }
  
  if (setsockopt(wt_fd, SOL_SOCKET, SO_REUSEADDR, &tmp, sizeof(tmp)) < 0)
    {
      perror("SO_REUSEADDR");
      goto error_out;
    }
  
  local.sin_family = AF_INET;
  local.sin_addr.s_addr = INADDR_ANY;
  local.sin_port = htons(WESTELL_UDP_NOTIFY_PORT);
  
  if (bind(wt_fd, (struct sockaddr *)&local, sizeof(local)) < 0)
    {
      perror("bind");
      goto error_out;
    }
  
  inet_aton(WESTELL_MULITCAST_GROUP, &mcast.imr_multiaddr);
  mcast.imr_address.s_addr   = INADDR_ANY;
  mcast.imr_ifindex          = ifindex;
  
  if (setsockopt(wt_fd, SOL_IP, IP_ADD_MEMBERSHIP, &mcast, sizeof(mcast)) < 0)
    {
      perror("IP_ADD_MEMBERSHIP");
      goto error_out;
    }
  
  /* setup tcp listen */
  listen_fd = socket(PF_INET, SOCK_STREAM, 0);
  if (listen_fd < 0)
    {
      perror("socket");
      goto error_out_drop;
    }
  
  {
    char *listen_ip = strdup(opt_serve);
    char *listen_port;
    
    listen_port = strchr(listen_ip, ':');
    
    if (!listen_port)
      {
        fprintf(stderr,"Invalid serve IP address: %s\n", listen_ip);
        free(listen_ip);
        goto error_out_drop;
      }
    
    *listen_port = '\0';
    listen_port++;
    
    local.sin_family = AF_INET;
    if (!inet_aton(listen_ip, &local.sin_addr))
      {
        fprintf(stderr,"Invalid serve IP address: %s\n", listen_ip);
        free(listen_ip);
        goto error_out_drop;
      }
    local.sin_port = atoi(listen_port);
    local.sin_port = htons(local.sin_port);
    free(listen_ip);
  }
  
  if (bind(listen_fd, (struct sockaddr *)&local, sizeof(local)) < 0)
    {
      perror("bind");
      goto error_out_drop;
    }
  
  if (listen(listen_fd, 10) < 0)
    {
      perror("listen");
      goto error_out_drop;
    }
  
  /* daemon? */
  if (opt_daemon)
    {
      pid_t pid;
      
      close(fileno(stdin));
      if (open("/dev/null", O_RDONLY) != fileno(stdin))
        {
          perror("open");
          goto error_out_drop;
        }
      
      pid = fork();
      if (pid < 0)
        {
          perror("fork 1");
          goto error_out_drop;
        }
      else if (pid > 0)
        {
          exit(0);
        }
      
      pid = setsid();
      if (pid < 0)
        {
          perror("setsid");
          goto error_out_drop;
        }
      
      pid = fork();
      if (pid < 0)
        {
          perror("fork 2");
          goto error_out_drop;
        }
      else if (pid > 0)
        {
          exit(0);
        }
    }
  
  /* get packet */
  
  if (verbose) printf("Waiting for Packet\n");
  
  while (1)
    {
      fd_set readset;
      FD_ZERO(&readset);
      FD_SET(wt_fd, &readset);
      FD_SET(listen_fd, &readset);
      
      if (select(wt_fd > listen_fd ? wt_fd+1 : listen_fd+1,
                 &readset, NULL, NULL, NULL) < 0)
        {
          if (errno == EINTR)
            {
              if (got_sigint)
                {
                  fprintf(stderr,"Interrupted\n");
                  goto error_out_drop;
                }
              continue;
            }
          else
            {
              perror("select");
              goto error_out_drop;
            }
        }
      
      if (FD_ISSET(wt_fd, &readset))
        {
          buffer_m[0] = 0x00;
          buffer_m[1] = 0x32;
          buffer_m[2] = 0x01;
          
          recv_len = BUFFER_SIZE;
          if (do_transaction(wt_fd, NULL, NULL, 0, buffer_r, &recv_len, buffer_m, 3) < 0)
            {
              goto error_out_drop;
            }
          process_notification(opt_rate, buffer_r, recv_len, serve_notification);
        }
      
      if (FD_ISSET(listen_fd, &readset))
        {
          struct sockaddr_in remote;
          socklen_t remote_len = sizeof(remote);
          int client_fd;
          int client;
          
          client_fd = accept(listen_fd, (struct sockaddr*)&remote, &remote_len);
          if (client_fd < 0)
            {
              perror("accept");
              goto error_out_drop;
            }
          
          for(client=0; client<MAX_CLIENTS; client++)
            {
              if (client_fds[client] == -1)
                {
                  if (verbose) printf("Accepted client from %s:%d on slot %d\n",
                                      inet_ntoa(remote.sin_addr),
                                      ntohs(remote.sin_port),
                                      client);
                  client_fds[client] = client_fd;
                  break;
                }
            }
          if (client == MAX_CLIENTS)
            {
              fprintf(stderr,"Too many clients, rejecting connection from %s:%d\n",
                      inet_ntoa(remote.sin_addr),
                      ntohs(remote.sin_port));
              close(client_fd);
            }
        }
    }
  retval = 0;
  
 error_out_drop:
  inet_aton(WESTELL_MULITCAST_GROUP, &mcast.imr_multiaddr);
  mcast.imr_address.s_addr   = INADDR_ANY;
  mcast.imr_ifindex          = ifindex;
  
  if (setsockopt(wt_fd, SOL_IP, IP_DROP_MEMBERSHIP, &mcast, sizeof(mcast)) < 0)
    {
      perror("IP_DROP_MEMBERSHIP");
      retval = -1;
      goto error_out;
    }
  
 error_out:
  close(wt_fd);
  if (listen_fd != -1)
    {
      close(listen_fd);
    }
  return retval;
}


static int do_disable(struct sockaddr_in *opt_bcast)
{
  int fd;
  int tmp = 1;
  ssize_t recv_len;
  
  fd = socket(PF_INET, SOCK_DGRAM, 0);
  if (fd < 0)
    {
      perror("socket");
      return -1;
    }
  
  if (setsockopt(fd, SOL_SOCKET, SO_BROADCAST, &tmp, sizeof(tmp)) < 0)
    {
      perror("SO_BROADCAST");
      goto error_out;
    }
  
  if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &tmp, sizeof(tmp)) < 0)
    {
      perror("SO_REUSEADDR");
      goto error_out;
    }
  
  opt_bcast->sin_port = htons(WESTELL_UDP_NOTIFY_PORT);
  
  /* get serial and model numbers */
  
  printf("Disabling Notification\n");
  
  buffer_s[0] = 0x00;
  buffer_s[1] = 0x30;
  buffer_s[2] = 0x00;
  
  recv_len = BUFFER_SIZE;
  if (do_transaction(fd, opt_bcast, buffer_s, 3, buffer_r, &recv_len, NULL, 0) < 0)
    {
      goto error_out;
    }

  printf("Disabled\n");
  
  close(fd);
  return 0;
  
 error_out:
  close(fd);
  return -1;
}



int main (int argc, char *argv[])
{
  int option_index = 0;
  int c;
  
  int opt_info = 0;
  int opt_enable = 0;
  int opt_disable = 0;
  int opt_monitor = 0;
  int opt_daemon = 0;
  const char *opt_serve = NULL;
  const char *opt_iface = NULL;
  struct sockaddr_in opt_bcast = {};
  unsigned int opt_rate = 1;
  int ifindex = 0;
  
  assert(sizeof(struct wt_notify_data_t) == DATA_TOTAL_SIZE);  

  optind = 0;
  while ((c = getopt_long(argc, argv, "", westell_long_options, &option_index)) >= 0)
    {
      switch (c)
        {
        case 0:
          break;
          
        case OPT_VERBOSE:
          verbose++;
          break;
          
        case OPT_ACTION_INFO:
          opt_info = 1;
          break;
          
        case OPT_ACTION_ENABLE:
          opt_enable = 1;
          break;
          
        case OPT_ACTION_DISABLE:
          opt_disable = 1;
          break;
          
        case OPT_ACTION_MONITOR:
          opt_monitor = 1;
          break;
          
        case OPT_ACTION_SERVE:
          opt_serve = optarg;
          break;
          
        case OPT_DAEMON:
          opt_daemon = 1;
          break;
          
        case OPT_DATEFMT:
          datefmt = optarg;
          break;
          
        case OPT_IFS:
          ifs = optarg;
          break;
          
        case OPT_RATE:
          opt_rate = strtoul(optarg, NULL, 0);
          break;
          
        case OPT_BCAST:
          if (!inet_aton(optarg, &opt_bcast.sin_addr))
            {
              usage(argv[0]);
            }
          break;
          
        case OPT_IFACE:
          opt_iface = optarg;
          break;
          
        default:
          usage(argv[0]);
        }
    }
  
  if (optind < argc)
    {
      usage(argv[0]);
    }
  
  if (opt_rate < 1)
    {
      opt_rate = 1;
    }
  
  if (opt_iface)
    {
      int fd;
      struct ifreq ifr = {};
      strncpy(ifr.ifr_name, opt_iface, IFNAMSIZ);

      fd = socket(PF_INET, SOCK_DGRAM, 0);
      if (fd < 0)
        {
          perror("socket");
          exit(1);
        }
      
      if (ioctl(fd, SIOCGIFINDEX, &ifr) < 0)
        {
          perror(opt_iface);
          exit(1);
        }
      
      ifindex = ifr.ifr_ifindex;
    }
  
  if (opt_monitor)
    {
      if ((opt_bcast.sin_addr.s_addr != INADDR_ANY) ||
          opt_serve ||
          opt_info ||
          opt_enable ||
          opt_disable ||
          opt_daemon)
        {
          usage(argv[0]);
        }
    }
  else if (opt_serve)
    {
      if ((opt_bcast.sin_addr.s_addr != INADDR_ANY) ||
          opt_monitor ||
          opt_info ||
          opt_enable ||
          opt_disable)
        {
          usage(argv[0]);
        }
    }
  else if (!opt_info && !opt_enable && !opt_disable)
    {
      usage(argv[0]);
    }
  else if ((opt_bcast.sin_addr.s_addr == INADDR_ANY) ||
           ifindex ||
           opt_daemon)
    {
      usage(argv[0]);
    }
  
  
  {
    struct sigaction sa = {};
    
    sa.sa_sigaction = westell_signal_handler;
    sa.sa_flags = SA_SIGINFO;
    sigfillset(&sa.sa_mask);
    sigaction(SIGALRM, &sa, NULL);
    sigaction(SIGINT, &sa, NULL);
    sigaction(SIGPIPE, &sa, NULL);
  }
  
  if (opt_info && (do_info(&opt_bcast) < 0))
    {
      exit(1);
    }
  
  if (opt_enable && (do_enable(&opt_bcast) < 0))
    {
      exit(1);
    }
  
  if (opt_disable && (do_disable(&opt_bcast) < 0))
    {
      exit(1);
    }
  
  if (opt_monitor && (do_monitor(ifindex, opt_rate) < 0))
    {
      exit(1);
    }
  
  if (opt_serve && (do_serve(ifindex, opt_rate, opt_daemon, opt_serve) < 0))
    {
      exit(1);
    }
  
  exit(0);
}
